@@ -1,8 +1,8 @@
 #
-#   o_dbm.rb - ֥Ȼظǡ١Object Base dbm
-#   	$Release Version: 0.3.1$
-#   	$Revision: 1.3 $
-#   	$Date: 1998/03/29 17:09:12 $
+#   o_dbm.rb - ֥Ȼظǡ١Object Base DBM
+#   	$Release Version: 0.5.1$
+#   	$Revision: 1.9 $
+#   	$Date: 2002/07/12 04:46:28 $
 #   	by Keiju ISHITSUKA(Nippon Rational Inc.)
 #
 # --
@@ -13,10 +13,10 @@
 require "e2mmap"
 
 class ObjectDBM
-  @RELEASE_VERSION = "0.3.1"
-  @LAST_UPDATE_DATE = "01/03/19"
+  @RELEASE_VERSION = "0.5.1"
+  @LAST_UPDATE_DATE = "02/07/12"
 
-  @RCS_ID='-$Id: o_dbm.rb,v 1.3 1998/03/29 17:09:12 keiju Exp keiju $-'
+  @RCS_ID='-$Id: o_dbm.rb,v 1.9 2002/07/12 04:46:28 keiju Exp keiju $-'
 
   extend Exception2MessageMapper
 
@@ -145,8 +145,7 @@ class ObjectDBM
     obj = @read_cache[key]
     return obj unless obj == NULL
       
-    return @default_value unless s = @db[key]
-    obj = Marshal.load(s)
+    return @default_value unless obj = @db[key]
     @read_cache[key] = obj if mode != NO_CACHING
     obj
   end
@@ -155,8 +154,9 @@ class ObjectDBM
     return self[key] = obj unless obj == :NO_OPTION
 
     return @write_cache[key] = obj unless (obj = @read_cache[key]) == NULL
-    return @default_value unless s = @db[key]
-    obj = @write_cache[key] = @read_cache[key] = Marshal.load(s)
+    return @default_value unless obj = @db[key]
+    @write_cache[key] = @read_cache[key] = obj
+
     @delete_cache.delete(key)
     obj
   end
@@ -181,9 +181,9 @@ class ObjectDBM
 
   def root_names
     keys = []
-    each_keys do
+    each_key do
       |key|
-      keys.push = keys
+      keys.push keys
     end
     keys
   end
@@ -191,7 +191,7 @@ class ObjectDBM
 
   def size
     no = 0
-    each_keys do
+    each_key do
       |key|
       no += 1
     end
@@ -221,18 +221,18 @@ class ObjectDBM
   def has_root_name?(root_name, mode = SCAN_DB)
     return true if @read_cache.key?(root_name)
     return false if mode == SCAN_CACHE_ONLY
-    @db.has_key?
+    @db.has_key?(root_name)
   end
   alias root_name? has_root_name?
   alias include? has_root_name?
   alias key? has_root_name?
 
   def has_root?(root, mode = SCAN_DB)
-    return has_root(root, mode){|x,y| x.eq?(y)} if iterator?
+    return has_root(root, mode){|x,y| x.equal?(y)} if iterator?
 
     if mode != SCAN_DB_ONLY
       @read_cache.each_value{|r| return true if yield root, r}
-      return false if mode == SCAN_CASHE_ONLY
+      return false if mode == SCAN_CACHE_ONLY
     end
     @db.each_value{|r| return true if yield root, r}
     false
@@ -261,10 +261,8 @@ class ObjectDBM
     end
 
     if mode != SCAN_CACHE_ONLY
-      @db.each do |key, row_v|
+      @db.each do |key, obj|
 	next unless mode == SCAN_DB_ONLY or @read_cache[key] == NULL
-
-	obj = Marshal.load(row_v) 
 	@read_cache[key] = obj if mode == READ_CACHING
 	@write_cache[key] = obj if mode == UPDATE
 	yield key, obj
@@ -286,7 +284,7 @@ class ObjectDBM
     if mode != SCAN_CACHE_ONLY
       @db.each_key do
 	|key|
-	yield key unless mode == SCANL_DB_ONLY or @read_cache[key] == NULL
+	yield key if mode == SCAN_DB_ONLY or @read_cache[key] == NULL
       end
     end
   end
@@ -327,8 +325,11 @@ class ObjectDBM
       end
       @transaction_mode = STATIC_TRANSACTION_MODE
       @current_transaction = StaticTransaction.new(self, mode, outer)
-      @current_transaction.transaction do
-	yield @current_transaction
+#      @current_transaction.transaction do
+#	yield @current_transaction
+#      end
+      @current_transaction.transaction do |txn|
+	yield txn
       end
     else
       if @transaction_mode == STATIC_TRANSACTION_MODE
@@ -368,8 +369,9 @@ class ObjectDBM
       txn.outer.write_cache = txn.write_cache.dup
       txn.outer.delete_cache = txn.delete_cache.dup
     else
-      close(HOLD_READ_CACHE)
-      @db = @db_adapter.open(@db_name)
+#      close(HOLD_READ_CACHE)
+#      @db = @db_adapter.open(@db_name)
+      flush_db
     end
   end
   
@@ -416,8 +418,9 @@ class ObjectDBM
     end
     
     for key, value in @write_cache
-      @db[key] = Marshal.dump(value)
+      @db[key] = value
     end
+    @db.flush
     @write_cache.clear
   end
   private :flush_db
@@ -511,7 +514,8 @@ class ObjectDBM
     def transaction
       @status = START
       begin
-	txn, value = catch(ABORT_LABEL){[nil, yield(@current_transaction)]}
+#	txn, value = catch(ABORT_LABEL){[nil, yield(@current_transaction)]}
+	txn, value = catch(ABORT_LABEL){[nil, yield(self)]}
 	if txn
 	  @status = ABORTING
 	  unless txn.equal?(self)
@@ -552,7 +556,7 @@ class ObjectDBM
     def checkpoint
       case @status
       when START
-	@odbm.flush(txn)
+	@odbm.flush(self)
       when COMMITED, ABORTED
 	Transaction.Fail ErrClosedTransaction
       end
@@ -626,7 +630,11 @@ class ObjectDBM
   class DB_Adapter
     # open database named <name>
     def self.open(name)
-      ODBM.Fail ErrAdapterInterfaceNotImplement, "open"
+      new(name)
+    end
+
+    def initialize(name)
+      ODBM.Fail ErrAdapterInterfaceNotImplement, "initialize"
       #@db
     end
 
@@ -637,16 +645,20 @@ class ObjectDBM
 
     # store value with <key>
     def []=(key, value)
-      ODBM.Fail ErrAdapterInterfaceNotImplement, "[]"
+      ODBM.Fail ErrAdapterInterfaceNotImplement, "[]="
     end
 
+    # testing for which the db have a key <key>
     def has_key?(key)
       @db.each_key do
 	|k|
 	return true if k == key
       end
     end
-
+    alias key? has_key?
+    alias include? has_key?
+    
+    # access all assoc in database.
     def each(&block)
       @db.each_key{|key|yield key, @db[key]}
     end
@@ -656,6 +668,7 @@ class ObjectDBM
       ODBM.Fail ErrAdapterInterfaceNotImplement, "each_key"
     end
 
+    # access all values in database.
     def each_value(&block)
       @db.each_key{|key|yield @db[key]}
     end
@@ -665,92 +678,186 @@ class ObjectDBM
       ODBM.Fail ErrAdapterInterfaceNotImplement, "delete"
     end
 
+    # flush database
+    def flush
+      ODBM.Fail ErrAdapterInterfaceNotImplement, "flush"
+    end
+
     # close database
     def close
       ODBM.Fail ErrAdapterInterfaceNotImplement, "close"
     end
   end
 
+  module HashLikeInterface
+    
+    def db
+      ODBM.Fail ErrAdapterInterfaceNotImplement, "db"
+    end
+
+    def [](key)
+      materialize_value(db[key])
+    end
+
+    def []=(key, value)
+      db[key] = serialize_value(value)
+    end
+
+    def has_key?(key)
+      db.key?(key)
+    end
+
+    def each
+      db.each{|k, v| yield k, materialize_value(v)}
+    end
+
+    def each_key(&block)
+      db.each_key &block
+    end
+
+    def each_value
+      db.each_value{|v| yield materialize_value(v)}
+    end
+
+    def delete(key)
+      db.delete(key)
+    end
+
+    def serialize_value(v)
+      ODBM.Fail ErrAdapterInterfaceNotImplement, "serialize_value(v)"
+    end
+    def materialize_value(v)
+      ODBM.Fail ErrAdapterInterfaceNotImplement, "materialize_value(v)"
+    end
+  end
+
   autoload :DBM, "dbm"
   class DBM_Adapter<DB_Adapter
-    def self.open(name)
+    
+    include HashLikeInterface
+
+    def initialize(name)
       @db = DBM.open(name)
     end
 
-    def [](key)
-      @db[key]
+    def db
+      @db
     end
 
-    def []=(key, value)
-      @db[key] = valuep
+    def serialize_value(v)
+      Marshal.dump(v)
     end
 
-    def key?(key)
-      @db.key?(key)
+    def materialize_value(v)
+      return v unless v
+      Marshal.load(v)
     end
 
-    def each(&block)
-      @db.each &block
-    end
-
-    def each_key(&block)
-      @db.each_key &block
-    end
-
-    def each_value(&block)
-      @db.each_value &block
+    def flush
+      # noop
     end
 
     def close
+      flush
       @db.close
       @db = nil
     end
+
   end
 
   autoload :GDBM, "gdbm"
   class GDBM_Adapter<DBM_Adapter
-    def self.open(name)
+    def initialize(name)
       @db = GDBM.open(name)
     end
+
+    def flush
+      @db.sync
+    end
   end
 
   autoload :SDBM, "sdbm"
   class SDBM_Adapter<DBM_Adapter
-    def self.open(name)
+    def initialize(name)
       @db = SDBM.open(name)
     end
   end
 
   class PHash_Adapter<DB_Adapter
-    def self.open(name)
+    include HashLikeInterface
+
+    def initialize(name)
       @db_name = name
-      @db = File.open(name)
-      ODBM.Fail ErrAdapterInterfaceNotImplement, "open"
+      if File.exist?(name)
+        file = File::open(name, 'r')
+	begin
+	  @hash = Marshal.load(file)
+	ensure
+	  file.close
+	end
+      else
+        @hash = {}
+      end
     end
 
-    # restore value with <key> 
-    def [](key)
-      ODBM.Fail ErrAdapterInterfaceNotImplement, "[]"
+    def db
+      @hash
+    end
+    def serialize_value(v)
+      v
     end
 
-    # store value with <key>
-    def []=(key, value)
-      ODBM.Fail ErrAdapterInterfaceNotImplement, "[]"
+    def materialize_value(v)
+      v
     end
 
-    # access all keys in database.
-    def each_key(&block)
-      ODBM.Fail ErrAdapterInterfaceNotImplement, "each_key"
-    end
-
-    # delete value with <key>
-    def delete(key)
-      ODBM.Fail ErrAdapterInterfaceNotImplement, "delete"
+    # commit database
+    def flush
+      newfile = @db_name + '.new'
+      file = open(newfile, 'w')
+      Marshal.dump(@hash, file)
+      file.close
+      File.rename(newfile, @db_name)
     end
 
     # close database
     def close
-      ODBM.Fail ErrAdapterInterfaceNotImplement, "close"
+      flush
+      @hash = nil
+    end
+  end
+
+  class VDB_Adapter<DB_Adapter
+    include HashLikeInterface
+
+    VDBS = {}
+
+    def initialize(name)
+      @db_name = name
+      @hash = VDBS[@db_name]
+      @hash = {} unless @hash
+    end
+
+    def db
+      @hash
+    end
+    def serialize_value(v)
+      v
+    end
+
+    def materialize_value(v)
+      v
+    end
+
+    # commit database
+    def flush
+      VDBS[@db_name] =  @hash
+    end
+
+    # close database
+    def close
+      flush
+      @hash = nil
     end
   end
 end
